#!/usr/local/bin/tclsh
# -*- tcl -*-
#
# Tcl Debugger Shell for Komodo
# - Interface between Komodo and the TclPro debugger engine
#
# Copyright (c) 2001 ActiveState Corporation
#
# Accepted options
# 	-port		int	Port we are commanded to serve on.
# 	-port-komodo	int	Port komodo listens on for connect back
#	-host-komodo	string	Name of host running Komodo
#	-cookie		string	Magic value 'protecting' the connection.
#	-file		path	Script to debug
#	-pause		boolean	Wait after end of debugging
#

catch {wm withdraw .} ; # just in case if wish is used as shell

package require log     ; # tcllib

set ::timing 0
set logtofile 0
#log::lvSuppress debug
log::lvSuppressLE warning ; # supress everything which is not an error
# critical ; # == suppress everything/all

#######################################################
# Extend logger module with custom procedure, setup logging

proc ::log::mlog {level text {pfx {}}} {
    foreach l [split $text \n] {
	::log::log $level $pfx$l
    }
    return
}

if {$logtofile} {
    set logfile [file rootname [info script]].log
    set logchan [open $logfile w]
    fconfigure $logchan -buffering none

    ::log::lvChannelForall $logchan
    ::log::log debug "... Logging into $logfile"
}

#######################################################
# Handle premature [exit]s => Log state and from where called.

rename ::exit ::ktdb::ExitCmd
proc ::exit {{state 0}} {
    ::log::log warning "exit $state"
    if {[info level] == 1} {
	::log::log warning "@  global level"
    } else {
	::log::log warning "by [info level -1]"
    }
    ::ktdb::ExitCmd $state
}

proc bgerror {args} {
    ::ktdb::dumpTclError
}

#######################################################

log::log debug "Proceeding into main procedure definitions"

#######################################################
# The code below lives in the "ktdb" namespace to prevent
# interference with the code to be debugged.
#
# ktdb = Komodo Tcl DeBugger.
#######################################################

namespace eval ::ktdb {
    variable  opts
    array set opts {}

    variable options {
	{port.arg        -1 {Port we have to listen on}}
	{port-komodo.arg -1 {Port Komodo is listening on for connect-back}}
	{host-komodo.arg {} {Name of host Komodo is running on}}
	{cookie.arg      {} {Cookie required for access}}
	{file.arg        {} {The file to debug}}
	{pause.arg        0 {}}
    }

    variable  resumeDisplay
    array set resumeDisplay {
	0 stop
	1 in
	2 over
	3 out
	4 go
    }
    variable  resumeCmd
    array set resumeCmd {
	0 -
	1 {dbg::step any}
	2 {dbg::step over}
	3 {dbg::step out}
	4 {dbg::step run}
    }

    variable eCodeV  .
    variable eInfoV  .
    variable codeV   .
    variable resultV .
    variable idV     .

    # Maintain mapping between tcl breakpoint ids and komodo
    # breakpoint numbers.

    variable  bp
    array set bp {}

    variable bpcount 0
    variable bpTemp {}

    # Constant strings for some replies.
    variable statusReply       ""
    variable contextNamesReply ""
    variable invalidBpReply    ""
    variable setDebugOkReply   ""
    variable virtualGlobalFrame ""

    # trigger variable for "resumeExecution". Written to when the
    # tclpro engine reports that it stopped and is waiting for more
    # commands.

    variable stepComplete

    # Filenames for virtual stackframes to handle specially.

    variable  virtual_file
    array set virtual_file {
	s,global komodo://tcl/debugger/script/global_frame
	f,komodo://tcl/debugger/script/global_frame global
    }
}

#######################################################
log::log debug main

proc ::ktdb::main {} {
    ::log::log debug "Coming up"

    if {[catch {
	if {$::timing} {
	    log::log info "TIME load pkg  [time ::load_required_packages 1]"
	    log::log info "TIME load eng  [time ::load_engine 1]"
	    log::log info "TIME proc opt  [time process_options 1]"
	    log::log info "TIME chk  opt  [time check_options 1]"
	    log::log info "TIME proc file [time process_file 1]"
	    log::log info "TIME setup srv [time setup_server 1]"
	} else {
	    ::load_required_packages
	    ::load_engine
	    process_options
	    check_options
	    process_file
	    setup_server
	}
	process_commands
    }]} {
	exitWithError 12 "-- Internal server error --"
    }
    exit 11
}

#######################################################
log::log debug exitWithError

proc ::ktdb::exitWithError {{state 0} {reason {}}} {
    dumpTclError $reason
    ::exit $state
}

#######################################################
log::log debug dumpTclError

proc ::ktdb::dumpTclError {{reason {}}} {
    global errorCode errorInfo

    if {![string equal $reason {}]} {
	log::log  critical "* $reason"
    }
    log::log  critical "* $errorCode"
    log::mlog critical $errorInfo "* "
    return
}

#######################################################
log::log debug load_required_packages

proc ::load_required_packages {} {
    log::log debug "Loading external packages ..."

    package require Tclx    ; # we want signal handling ...
    package require md5     ; # tcllib
    package require cmdline ; # tcllib
    package require dom
    package require SOAP
    package require XMLRPC
    package require SOAP::Utils
    catch {namespace import -force SOAP::Utils::*}

    foreach s {15 17 19} {
	if {[catch {signal trap [list $s] ::ktdb::propagate} msg]} {
	    log::log warning "trap trouble <$s> $msg"
	}
    }
    return
}

#######################################################
log::log debug load_engine

proc ::load_engine {} {
    log::log debug "Loading debugger engine ..."

    ## We assume that the tcl debugger backend is stored in a directory
    ## 'engine' below the directory containing this script. We also assume
    ## that the main entry file is named "debug_main.tcl".

    if {[catch {
	log::log debug "@ [file join [file dirname [info script]] engine debug_main.tcl]"
	uplevel #0 [list source [file join [file dirname [info script]] engine debug_main.tcl]]
    } reason]} {
	::ktdb::exitWithError 2 "-- engine not available, shutting down --"
    }
    return
}

#######################################################
log::log debug process_options

proc ::ktdb::process_options {} {
    global argv
    variable opts
    variable options

    ::log::log debug "* processing options ..."
    ::log::log debug "* $argv"

    if {[catch {array set opts [::cmdline::getoptions argv $options]} msg]} {
	exitWithError 3 "Option Error"
    }

    # Take the remaining options as the options to the debugged application.
    set opts(arguments) {}
    foreach arg $argv {
	switch -glob -- $arg {
	    -- {
		set argv [lrange $argv 1 end]
		break
	    }
	    * {
		lappend opts(arguments) [lindex $argv 0]
		set argv [lrange $argv 1 end]
	    }
	}
    }

    # port = port to serve on

    set   opts(port-tdb) $opts(port)
    unset opts(port)

    # Filling in missing values, if possible.

    if {$opts(port-komodo) != {}} {
	set opts(connectback) 1
    } else {
	set opts(connectback) 0
    }
    if {$opts(host-komodo) == ""} {
	set opts(host-komodo) localhost
    }
    if {$opts(cookie) == "" && $opts(connectback)} {
	set opts(cookie) [md5::md5 [clock clicks][pid]]
    }
    if {$opts(file) != ""} {
	set opts(file) [fresolve $opts(file)]
    }

    # display generated configuration

    foreach {key value} [array get opts] {
	::log::log debug "  $key\t= $value"
    }
}

#######################################################
log::log debug check_options

proc ::ktdb::check_options {} {
    variable opts

    ::log::log debug "* checking options ..."

    if {$opts(connectback)} {
	if {
	    $opts(host-komodo) == "" ||
	    $opts(port-komodo) == ""
	} {
	    log::log critical "Can't talk to komodo, host and/or port missing"
	    exit 4
	}
    } else {
	if {$opts(port-tdb) == ""} {
	    log::log critical "No port to serve on specified"
	    exit 5
	}
    }
    if {$opts(file) == ""} {
	log::log critical "File to debug missing"
	exit 6
    }
    if {$opts(cookie) == ""} {
	log::log critical "Cookie missing"
	exit 7
    }
    return
}

#######################################################
log::log debug setup_server

proc ::ktdb::setup_server {} {
    variable opts

    ::log::log debug "setting up the server ..."

    if {!$opts(connectback)} {
	# Take the port number from the cmdline and start server there.

	::log::log debug "* console environment ..."
	::log::log debug "  port   = $opts(port-tdb) ..."
	::log::log debug "  cookie = $opts(cookie) ..."

	socket -server ::ktdb::newConnection $opts(port-tdb)
    } else {
	# Connect back: Start a server (listening socket) on a random
	# port and transfer the port/cookie information to the client
	# (Komodo) using its listening port as found in our
	# configuration.

	::log::log debug "* komodo environment ..."

	set srv            [socket -server ::ktdb::newConnection 0]
	set opts(port-tdb) [lindex [fconfigure $srv -sockname] 2]

	::log::log debug "  port   = $opts(port-tdb) ..."
	::log::log debug "  cookie = $opts(cookie) ..."
	::log::log debug "  connecting back @ ($opts(host-komodo) $opts(port-komodo)) ..."

	set    client [socket $opts(host-komodo) $opts(port-komodo)]
	puts -nonewline $client "Tcl:$opts(port-tdb):$opts(cookie)"
	close $client
    }
    return
}

#######################################################
log::log debug process_file

proc ::ktdb::process_file {} {
    variable opts

    # process the chosen file, insert the hook
    # statements intercepted by the debugging engine to
    # stay in control of the execution.

    ::log::log debug "launching file to debug ..."

    # Initialize debugger first, then reach into its data structures
    # to set a trace upon the application state (I want to monitor the
    # state to know how it is changing in case some debugger API
    # command errors out with a message about called while in a wrong
    # state).

    ::debugger::init {} {}

    namespace eval ::dbg {
	variable appState dead
    }
    trace variable ::dbg::appState w ::ktdb::traceAppState

    # Next step is to create the listening port for the communicaton
    # between the debugger engine here and the nub in the launched
    # application there.

    ::dbg::setServerPort random
    logSystemState

    ## Create the callbacks from the engine into the komodo interface,
    ## i.e. us. We have to react to soem events, and some callbacks
    ## are just to monitor and log the behaviour of the engine.

    dbg::register linebreak  {::ktdb::linebreakHandler}
    dbg::register varbreak   {::ktdb::varbreakHandler}
    dbg::register userbreak  {::ktdb::userbreakHandler}
    dbg::register exit       {::ktdb::exitHandler}
    dbg::register error      {::ktdb::errorHandler}
    dbg::register result     {::ktdb::resultHandler}
    dbg::register attach     {::ktdb::attachHandler}

    ###    dbg::register cmdresult  {gui::cmdresultHandler}
    ###    dbg::register instrument {gui::instrumentHandler}
    # Register the error handler for errors during instrumentation.
    ###    set instrument::errorHandler gui::instrumentErrorHandler

    ## Now we can launch the tcl shell with the application to debug.
    ## We will be notified via 'attachHandler' when it connects back
    ## to us.

    if {[catch {
	::dbg::start [info nameofexecutable] {} $opts(file) $opts(arguments) ktdbg[pid]
    }]} {
	exitWithError 14 "Application Initialization Error"
    }

    # Wait for the nub to attach before proceeding with our initialization ...
    # Danger: If the nub never connects this application will stay
    # here indefinitely and Komodo will never be able to talk to it.

    set   ::forward 0
    vwait ::forward

    ## Instrument the file now, so that incoming addBreakpoint
    ## requests work even before we actually step into the
    ## application. We don't care for the result here, just that the
    ## block is defined, instrumented and thus knowing the valid bp lines.

    # note that the instrumentor is fully initialized only after the
    # nub attached to the engine. Before that happens it doesn't know
    # the version of tcl to debug and thus doesn't load the parser
    # definitions.

    set fname    [fresolve $opts(file)]
    dbg::Instrument $fname [read [set fh [open $fname r]]][close $fh]

    return
}

#######################################################
log::log debug process_commands

proc ::ktdb::process_commands {} {
    # Wait for xmlrpc requests made by the client.

    ::log::log debug "entering event loop ..."

    vwait forever
    return -code error "Illegal jump out of event loop"
}

#######################################################
log::log debug newConnection

proc ::ktdb::newConnection {sock host port} {
    ::log::log debug "new connection @ $sock ($host, $port) ..."

    fconfigure $sock -blocking 1 -translation binary

    if {$::timing} {
	fileevent $sock readable [list ::ktdb::newRequestTime $sock $host $port]
    } else {
	fileevent $sock readable [list ::ktdb::newRequest $sock $host $port]
    }
    return
}

#######################################################
log::log debug newRequest

proc ::ktdb::newRequestTime {sock host port} {
    log::log debug "TIME handle rq [time {::ktdb::newRequestTimed $sock $host $port} 1]"
}

proc ::ktdb::newRequest {sock host port} {
    variable opts

    if {[eof $sock]} {
	# Lost connection ... Kill server ?

	::log::log debug "lost connection, exiting ..."
	close $sock
	catch {dbg::quit}
	#catch {kill $::dbg::appPid}
	exit 0
    }

    # Retrieve headers, decode them, perform minimal checks, read and
    # process the xml following it ...

    set request {}
    set line .
    while {$line != {} && ! [eof $sock]} {
	gets $sock line
	lappend request [set line [string trim $line \r]]

	::log::log debug ">>> $line"
    }

    #puts "[join $request \n]"

    set http_request [split [lindex $request 0]] ; # 1st line, split into fields
    set http_action  [lindex $http_request 0]  ;# type of request
    set http_url     [lindex $http_request 1]  ;# what URL requested

    ::log::log debug "$http_request"

    if {[string tolower $http_action] != "post"} {
	requestError $sock "Invalid action $http_action"
	return
    }

    set clength_line [lsearch -regexp $request {^Content-Length:}]

    if { $clength_line < 0 } {
	requestError $sock  "Content-Length: missing"
	return
    }

    set clength [string trim [lindex [split [lindex $request $clength_line] :] 1]]

    if {![string is integer $clength]} {
	requestError $sock "Content-Length: positive integer number expected"
	return
    } elseif {$clength < 1} {
	requestError $sock "Content-Length: positive integer number expected"
	return
    }

    ::log::log debug "=== $clength bytes to read ..."

    set xml [read $sock $clength]
    set xml [string map [list "\r\n" "\n"] $xml]
    ::log::mlog debug $xml ">>> "

    # Parse the incoming request, extract method and arguments,
    # dispatch to handler procedure, send back a reply.

    set calldata [xmlrpc_parse_request $xml]

    foreach {methodName argValues} $calldata {break} ; # lassign
    set argValues [lindex $argValues 0]

    log::log info "method:    $methodName"
    log::log info "arguments: $argValues"

    array set data $argValues

    # Check the cookie

    if {0 != [string compare $data(cookie) $opts(cookie)]} {
	requestError $sock "Access denied"
	return
    }

    if {[info commands protocol/$methodName] == {}} {
	requestError $sock "Unknown protocol request $methodName"
	return
    }

    ## The request is ok, formally. Additional checks can only be done
    ## inside of the methods. Launch the method and write the reply
    ## back to the client.

    ###  NOTE
    ###
    #### Komodo has a synchronous API. Check the state of the tclpro
    #### engine and do not process the request until it has stopped.

    switch -exact -- $dbg::appState {
	stopped {
	    # Hurry up, the engine is waiting for our requests
	}
	running {
	    # Wait for the engine to stop before executing the incoming request.
	    log::log notice "Delaying request, waiting for engine to stop"
	    vwait ::ktdb::stepComplete
	}
	dead {
	    # The engine is dead, so there is no point in executing the incoming
	    # requests anymore. Reply with a fault. Usually. Some messages are
	    # still allowed, like hello.

	    if {$methodName != "Requester.hello"} {
		requestError $sock "Engine is dead, unable to process requests"
		return
	    } 
	}
    }

    if {[catch {
	xmlrpc_wrap_xml [xmlrpc_gen_reply [protocol/$methodName $argValues]]
    } reply]} {
	dumpTclError
	requestError $sock $reply
    } else {
	puts  $sock $reply
	flush $sock
    }

    close $sock
    return
}

proc ::ktdb::newRequestTimed {sock host port} {
    variable opts

    if {[eof $sock]} {
	# Lost connection ... Kill server ?

	::log::log debug "lost connection, exiting ..."
	close $sock
	catch {dbg::quit}
	#catch {kill $::dbg::appPid}
	exit 0
    }

    # Retrieve headers, decode them, perform minimal checks, read and
    # process the xml following it ...

    set cycles [time {
	set request {}
	set line .
	while {$line != {} && ! [eof $sock]} {
	    gets $sock line
	    lappend request [set line [string trim $line \r]]
	    
	    ::log::log debug ">>> $line"
	}

	#puts "[join $request \n]"

	set http_request [split [lindex $request 0]] ; # 1st line, split into fields
	set http_action  [lindex $http_request 0]  ;# type of request
	set http_url     [lindex $http_request 1]  ;# what URL requested

	::log::log debug "$http_request"

	if {[string tolower $http_action] != "post"} {
	    requestError $sock "Invalid action $http_action"
	    return
	}

	set clength_line [lsearch -regexp $request {^Content-Length:}]

	if { $clength_line < 0 } {
	    requestError $sock  "Content-Length: missing"
	    return
	}

	set clength [string trim [lindex [split [lindex $request $clength_line] :] 1]]

	if {![string is integer $clength]} {
	    requestError $sock "Content-Length: positive integer number expected"
	    return
	} elseif {$clength < 1} {
	    requestError $sock "Content-Length: positive integer number expected"
	    return
	}

	::log::log debug "=== $clength bytes to read ..."

	set xml [read $sock $clength]
	set xml [string map [list "\r\n" "\n"] $xml]
	::log::mlog debug $xml ">>> "
    } 1] ; # {}
    log::log debug "TIME html read $cycles"

    set cycles [time {
	# Parse the incoming request, extract method and arguments,
	# dispatch to handler procedure, send back a reply.

	set calldata [xmlrpc_parse_request $xml]

	foreach {methodName argValues} $calldata {break} ; # lassign
	set argValues [lindex $argValues 0]

	log::log info "method:    $methodName"
	log::log info "arguments: $argValues"
    } 1] ; # {}
    log::log debug "TIME xml decod $cycles"

    array set data $argValues

    # Check the cookie

    if {0 != [string compare $data(cookie) $opts(cookie)]} {
	requestError $sock "Access denied"
	return
    }

    if {[info commands protocol/$methodName] == {}} {
	requestError $sock "Unknown protocol request $methodName"
	return
    }

    ## The request is ok, formally. Additional checks can only be done
    ## inside of the methods. Launch the method and write the reply
    ## back to the client.

    ###  NOTE
    ###
    #### Komodo has a synchronous API. Check the state of the tclpro
    #### engine and do not process the request until it has stopped.

    switch -exact -- $dbg::appState {
	stopped {
	    # Hurry up, the engine is waiting for our requests
	}
	running {
	    # Wait for the engine to stop before executing the incoming request.
	    log::log notice "Delaying request, waiting for engine to stop"
	    vwait ::ktdb::stepComplete
	}
	dead {
	    # The engine is dead, so there is no point in executing the incoming
	    # requests anymore. Reply with a fault. Usually. Some messages are
	    # still allowed, like hello.

	    if {$methodName != "Requester.hello"} {
		requestError $sock "Engine is dead, unable to process requests"
		return
	    } 
	}
    }

    set cycles [time {
	if {[catch {
	    xmlrpc_wrap_xml [xmlrpc_gen_reply [protocol/$methodName $argValues]]
	} reply]} {
	    dumpTclError
	    requestError $sock $reply
	} else {
	    puts  $sock $reply
	    flush $sock
	}
    } 1] ; # {}
    log::log debug "TIME rq hdl/rp $cycles"

    close $sock
    return
}

#######################################################
log::log debug requestError

proc ::ktdb::requestError {sock reason} {
    ::log::mlog error $reason "XMLRPC: "
    puts  $sock [xmlrpc_wrap_xml [XMLRPC::fault 500 $reason]]
    flush $sock
    close $sock
    return -code return
}

#######################################################
log::log debug do_encoding

proc ::ktdb::do_encoding {xml} {
    if {[binary scan $xml ccc c0 c1 c2] == 3} {
	if {$c0 == -1 && $c1 == -2} {
	    ::log::log debug "encoding: UTF-16 little endian"
	    set xml [encoding convertfrom unicode $xml]
	} elseif {$c0 == -2 && $c1 == -1} {
	    ::log::log debug "encoding: UTF-16 big endian"
	    binary scan $xml S* xml
	    set xml [encoding convertfrom unicode [binary format s* $xml]]
	} elseif {$c0 == -17 && $c1 == -69 && $c2 == -65} {
	    ::log::log debug "encoding: UTF-8"
	    set xml [encoding convertfrom utf-8 $xml]
	}
    }
    return $xml
}

#######################################################
log::log debug xmlrpc_parse_request

proc ::ktdb::xmlrpc_parse_request {xml} {

    regsub -all ">\[ \t\n\]*<" [do_encoding $xml] \000 xml
    set xml [split $xml \000]

    ## puts $xml

    array set arg {}
    set methodName {}
    set lastName {}
    foreach item $xml {
	## puts @$item
	if {[string first > $item] < 0} {continue}
	set code {}
	regexp {^([^>]*)>([^<]*)</.*$} $item -> code data
	switch -exact -- $code {
	    methodName {
		## puts "\t$code <$data>"
		set methodName $data
	    }
	    name {
		if {$lastName != {}} {
		    ## puts "\t$code $lastName <>"
		    set arg($lastName) {}
		}
		## puts "\t$code <$data>"
		set lastName $data
	    }
	    {} {}
	    default {
		## puts "\t$code $lastName <$data>"
		set arg($lastName) $data
		set lastName {}
	    }
	}
    }
    return [list $methodName [list [array get arg]]]
}

#######################################################
log::log debug xmlrpc_gen_reply

proc ::ktdb::xmlrpc_gen_reply {results} {
    log::log info "method returns \[$results\]"

    return <methodResponse><params><param>$results</param></params></methodResponse>
}

#######################################################
log::log debug xmlrpc_wrap_xml

proc ::ktdb::xmlrpc_wrap_xml {xml} {
    set head {HTTP/1.1 200 OK}
    set head [join [list $head \
            "Content-Type: text/xml" \
            "Content-Length: [string length $xml]"\
            "" ] "\n" ]

    set reply "${head}\n${xml}"
    ::log::mlog debug $reply "<<< "

    return $reply
}

#######################################################

proc ::ktdb::logSystemState {} {
    log::log debug "$dbg::appState [dbg::getServerPortStatus]"
}

#######################################################

proc ::ktdb::propagate {args} {
    # We are going down, kill the launched application too.
    log::log critical "Debugger terminated by signal, killing child"
    #catch {dbg::quit}
    catch {kill $::dbg::appPid}
    exit 0
}

#######################################################
#######################################################
#######################################################
###
### Protocol handlers -- Komodo requests
###

#######################################################

proc ::ktdb::protocol/Requester.brk {arguments} {
    # DO NOT IMPLEMENT, not part of current protocol
    return {}
}

#######################################################
# Redirection of internal output from the engine to a socket

proc ::ktdb::protocol/Requester.redirectStdout {arguments} {
    # Arguments = port/int
    global logchan
    variable opts

    array set param $arguments
    validateParam port int

    log::log debug "Redirecting stdout to ($opts(host-komodo)/$port)"

    if {[catch {socket $opts(host-komodo) $port} sock]} {
	log::log error $sock
	return -code error "Could not connect to port \"$port\""
    }
    # Remove the test connection
    close $sock

    # Closing a std channel causes tcl to designate the next openend
    # channel as that std channel

    close stdout
    socket $opts(host-komodo) $port
    fconfigure stdout -buffering none

    puts  stdout "Stdout is redirected ..."

    ## And from now on we log not into a file, but the socket.

    #::log::lvChannelForall stdout
    #::log::log debug "... Logging to ($opts(host-komodo)/$port)"
    #close $logchan

    return {}
}

#######################################################
# Redirection of internal output from the engine to a socket

proc ::ktdb::protocol/Requester.redirectStderr {arguments} {
    # Arguments = port/int

    variable opts

    array set param $arguments
    validateParam port int

    log::log debug "Redirecting stderr to ($opts(host-komodo)/$port)"

    if {[catch {socket $opts(host-komodo) $port} sock]} {
	log::log error $sock
	return -code error "Could not connect to port \"$port\""
    }
    # Remove the test connection
    close $sock

    # Closing a std channel causes tcl to designate the next openend
    # channel as that std channel

    close stderr
    socket $opts(host-komodo) $port
    fconfigure stderr -buffering none

    puts  stderr "Stderr is redirected ..."
    return {}
}

#######################################################
# Redirection of internal input of the engine to a socket

proc ::ktdb::protocol/Requester.redirectStdin {arguments} {
    # Arguments = port/int

    variable opts

    array set param $arguments
    validateParam port int

    log::log debug "Redirecting stdin to ($opts(host-komodo)/$port)"

    if {[catch {socket $opts(host-komodo) $port} sock]} {
	log::log error $sock
	return -code error "Could not connect to port \"$port\""
    }
    # Remove the test connection
    close $sock

    # Closing a std channel causes tcl to designate the next openend
    # channel as that std channel

    close stdin
    socket $opts(host-komodo) $port

    return {}
}

#######################################################
# Ping engine

proc ::ktdb::protocol/Requester.hello {arguments} {
    # Arguments = none
    variable statusReply
    return $statusReply
}

#######################################################
# Stepping through the code, run, next, next-in, over, out, stop

proc ::ktdb::protocol/Requester.resumeExecution {arguments} {
    # Arguments = request/int

    variable resumeDisplay
    variable resumeCmd

    array set param $arguments
    validateParam request int

    if {($request < 0) || (4 < $request)} {
	return -code error {Request out of range [0..4]}
    }

    log::log debug "Resume : $resumeDisplay($request)"
    logSystemState

    if {$request == 0} {
	# STOP = session terminates = kill and exit.
	dbg::quit
	after idle ::exit 0
	return {}
    }

    eval $resumeCmd($request)
    return {}
}

#######################################################
# Runtime stack, how deep.

proc ::ktdb::protocol/Requester.getNumFrames {arguments} {
    # Arguments = none
    # Return 0 if application is not running.

    # Query the engine about this.
    logSystemState

    set numFrames [llength [dbg::getStack]]

    return [xstruct numFrames [xint $numFrames]]
}

#######################################################
# Runtime stack, what is on it ? 0 = current frame

proc ::ktdb::protocol/Requester.getFrames {arguments} {
    # Arguments = none
    # Ignore 'arguments', even if something is set.

    logSystemState

    # Query engine and transform result into the expected form.
    # In the notation of Tcl 0 is the global frame, but Komodo
    # expects 0 as the id of the current frame. Reverse the list
    # holding the converted stack before using it to create the
    # final result.

    set stack [dbg::getStack]

    log::log debug $stack

    return [xstruct frames [xarray [lreverse [convertFrames $stack]]]]
}

#######################################################

proc ::ktdb::convertFrames {stack} {
    variable virtual_file

    set rstack [list]
    foreach frame $stack {
	log::log info "*- $frame"

	foreach {level location type} $frame break

	set lineno [loc::getLine $location]

	if {$lineno == {}} {
	    set lineno 1
	}

	switch -exact -- $type {
	    global {
		set file  $virtual_file(s,global)
		set where {global level}
		# Skip over the frame for the global level. Not relevant.
		# continue
	    }
	    configure -
	    debugger_eval -
	    class -
	    event -
	    uplevel -
	    method -
	    package -
	    namespace -
	    proc {
		# We don't set the file here. This means that the file
		# information from the previous levels is carried on to
		# down to the other frames.

		set where [string trim "$type [join [lrange $frame 3 end]]"]
	    }
	    source {
		set file [uriname [lindex $frame 3]] ; # Location is URI !
		set where source
	    }
	    default {
		return -code error "Unknown stack frame \"$type\""
	    }
	}

	lappend rstack [xstruct file [xstring $file] lineno [xint $lineno] where [xstring $where]]
    }

    return $rstack
}

#######################################################

proc ::lreverse {list} {
    set result [list]
    foreach item $list {
	set result [linsert $result 0 $item]
    }
    return $result
}

#######################################################

proc ::ktdb::protocol/Requester.getSourceCode {arguments} {
    # Arguments = filename/string

    variable virtual_file
    variable virtualGlobalFrame

    array set param $arguments
    validateParam filename string

    if {[info exists virtual_file(f,$filename)]} {
	set fcode $virtual_file(f,$filename)
	switch -exact -- $fcode {
	    global {
		# The global stackframe has no source code.
		return $virtualGlobalFrame
	    }
	}
    }

    set fname [realname $filename]
    return [xstruct source [xstring [blk::getSource [blk::makeBlock $fname]]]]
}

#######################################################
# Value of variables, expressions, namespaces ...

proc ::ktdb::protocol/Requester.getDebugPropertyInfo {arguments} {
    # Arguments = frame_index/int, expr/string, radix/string

    array set param $arguments
    validateParam frame_index int
    validateParam expr        string
    validateParam radix       string

    # First determine the true stack level the caller refers to.

    set tclframe [frame2tclframe $frame_index]

    # Now evaluate the expression and return the result. An error is a
    # displayable result too!
    #
    # Discern with what expression we are dealing:
    # 1. variable   => return contents
    # 2. namespace  => return empty value, but set children
    # 3. expression => run it through [expr].
    #
    # The canonical fullname of the expression is written to cexpr

    if {[catch {set etype [expressionType $tclframe $expr -> cexpr ename]} emsg]} {
	set name     $expr
	set fullname $expr
	set value    $emsg
	set type     Error
	set children 0
    } else {
	set fullname $cexpr
	set name     $cexpr
	set children 0
	switch -exact -- $etype {
	    scalar-variable {
		set children 0
		set type     string
		set cmd      [list set $ename]
	    }
	    array-variable {
		# At this point we know that the array does exist, i.e. we
		# are free to query the interpreter about its properties

		set children \
			[expr {[GetResult [dbg::evaluate $tclframe \
			[list array size $ename]]] > 0}]
		set type     array
		set cmd      {}
		set value    "\{...\}"
	    }
	    namespace {
		# At this point we know that the namespace exists and
		# are free to ask the interpreter about its properties.

		if {
		    ([llength [GetResult [dbg::evaluate 0 [list ::info vars "${ename}::*"]]]] > 0) ||
		    ([llength [GetResult [dbg::evaluate 0 [list ::namespace children $ename]]]] > 0)
		} {
		    set children 1
		}
		set type     namespace
		set cmd      {}
		set value    "\{...\}"
	    }
	    expression {
		set type     expression
		set cmd      [list expr $cexpr]
	    }
	}
	if {$cmd != {}} {
	    if {[catch {GetResult [dbg::evaluate $tclframe $cmd]} value]} {
		set type     Error
		set children 0
	    }
	}
    }
    return [xstruct propertyInfo [xproperty $name $fullname $value $type $children]]
}

proc ::ktdb::expressionType {tclframe expression -> evar enamevar} {
    # 1. Preliminary categorization by application of syntax rules.
    # 2. Additional checking in the chosen category to determine
    #    existence, possibly a finer categorization or a recategorization.

    upvar $evar     canonical
    upvar $enamevar ename
    set canonical $expression
    set ename     $expression

    # Check for variables
    #    "$" \[A-Za-z0-9_\]+            | variable, determine scalar/array
    #    "$" \[A-Za-z0-9_\]+ "(" .* ")" | scalar variable in array
    #    "$" bracel \[^bracer\] bracer  | variable, check for "()".
    #    "$" .*                         | illegal !

    if {[regexp {^\$([:A-Za-z0-9_]+)$} $expression -> name]} {
	log::log info "$expression => $name | some variable"
	set ename $name
	return [eTypeVar $tclframe $name]
    }
    if {[regexp {^\$(([:A-Za-z0-9_]+)\((.*)\))$} $expression -> name base index]} {
	log::log info "$expression => $name | some array item"
	set ename $name
	return [eTypeArrayItem $tclframe $name $base $index]
    }
    if {[regexp {^\$\{(.+)\}$} $expression -> name]} {
	if {[regexp {^([ :A-Za-z0-9_]+)$} $name]} {
	log::log info "$expression => $name | some variable /2"
	    set ename $name
	    return [eTypeVar $tclframe $name]
	}
	if {[regexp {^([ :A-Za-z0-9_]+)\((.*)\)$} $name -> base index]} {
	    log::log info "$expression => $name | some array item /2"
	    set ename $name
	    return [eTypeArrayItem $tclframe $name $base $index]
	}
	log::log info "$expression => $name | illegal syntax /2"
	return -code error "Illegal variable syntax"
    }

    if {[regexp {^\$.*$} $expression]} {
	# A dollar sign in front, but none of the legal variablee syntaxes ...
	log::log info "$expression => $name | illegal syntax"
	return -code error "Illegal variable syntax"
    }

    # Now check that the thing is in legal namespace syntax.
    # We allow spaces.

    if {[regexp {^[:A-Za-z0-9_ ]+$} $expression]} {
	# check that the namespace exists. If it does not check if the
	# specified expression is possibly the name of an variable. If
	# yes, assume that the user actually requested the
	# variable. If there is no variable for the name either return
	# an error about namespace/variable not found

	if {[catch {GetResult [dbg::evaluate $tclframe [list namespace children $expression]]} emsg]} {
	    # Given the syntax above this cannot be an array element

	    if {![GetResult [dbg::evaluate $tclframe [list info exists $expression]]]} {
		# Expression is neither namespace nor variable name.

		return -code error "unknown namespace/variable \"$expression\""
	    }

	    # Variable exists. Determine its type and switch the canonical
	    # representation. Use the '${}' notation to ensure that whitespace
	    # is quoted properly.

	    set canonical [varquote $expression]

	    if {[GetResult [dbg::evaluate $tclframe [list array exists $expression]]]} {
		return array-variable
	    } else {
		return scalar-variable
	    }
	}
	return namespace
    }

    # Neither variable nor namespace. Assume that it is an expression.

    return expression
}

proc ::ktdb::eTypeVar {tclframe name} {
    # Determine if existing and if array or scalar.

    if {![GetResult [dbg::evaluate $tclframe [list info exists $name]]]} {
	# emulate response of tcl interpreter if we had actually
	# tried to read the variable
	return -code error "can't read \"$name\": no such variable"
    }
    
    # Variable exists. Now determine its type.

    if {[GetResult [dbg::evaluate $tclframe [list array exists $name]]]} {
	return array-variable
    }

    return scalar-variable
}

proc ::ktdb::eTypeArrayItem {tclframe name base index} {
    # Determine if the variable exists

    if {![GetResult [dbg::evaluate $tclframe [list info exists $base]]]} {
	# emulate response of tcl interpreter if we had actually
	# tried to read the variable
	return -code error "can't read \"$name\": no such variable"
    }

    # Base exists, make sure that it is an array.

    if {![GetResult [dbg::evaluate $tclframe [list array exists $base]]]} {
	# emulate response of tcl interpreter if we had actually
	# tried to read the variable
	return -code error "can't read \"$name\": variable isn't array"
    }

    # Check that the key exists in the array

    if {![GetResult [dbg::evaluate $tclframe [list info exists $name]]]} {
	# emulate response of tcl interpreter if we had actually
	# tried to read the variable
	return -code error "can't read \"$name\": no such element in array"
    }

    return scalar-variable
}


#######################################################
# Set variables

proc ::ktdb::protocol/Requester.setDebugProperty {arguments} {
    # Arguments = frame_index/int, lhs/string, rhs/string, radix/string

    variable setDebugOkReply

    array set param $arguments
    validateParam frame_index int
    validateParam lhs         string
    validateParam rhs         string
    validateParam radix       string

    # First determine the true stack level the caller refers to.

    set tclframe [frame2tclframe $frame_index]

    # Now evaluate the expression. It might not be a true command but just a value.
    # Any string enclosed in [ and ] is interpreted as a command to evaluate.
    # To set the variable to the value of another variable use the "set" command.
    # I.e. "[set <othervar>]".

    if {[catch {set etype [expressionType $tclframe $lhs -> cexpr ename]} emsg]} {
	return [xstruct setResultString [xstring $emsg]]
    }
    switch -exact -- $etype {
	    namespace {
		return [xstruct setResultString [xstring "Cannot assign to namespace"]]
	    }
	    expression {
		return [xstruct setResultString [xstring "Cannot assign to expression"]]
	    }
	    array-variable {
		return [xstruct setResultString [xstring "Cannot assign to array variable"]]
	    }
    }

    # The left-handside we got, the fullName of the property, is
    # actually the command prefix one has to use for setting the
    # property (if the property is settable).

    set       cmd set
    lappend cmd $ename
    lappend cmd $rhs

    set fail [catch {GetResult [dbg::evaluate $tclframe $cmd]} msg]
    if {$fail} {
	return [xstruct setResultString [xstring $msg]]
    }

    return $setDebugOkReply
}

#######################################################
# Get variables, namespaces, ... for a structured object = namespace

proc ::ktdb::protocol/Requester.getChildrenInfos {arguments} {
    # Arguments = frame_index/int parent_expr/string, radix/int

    array set param $arguments
    validateParam context     string
    validateParam frame_index int
    validateParam parent_expr string
    validateParam radix       int

    # parent_expr is for us one of
    #
    # "${" varname "}" => Array 'varname'.
    # string           => Namespace 'string'.
    #
    # This allows us to distinguish between arrays and namespaces ...

    # Note: The "context" information can be empty (this signals a call out of the watch window).
    # In that case we use the current stack frame to evaluate the property.

    switch -exact -- $context {
	globals {
	    set tclframe 0
	}
	locals {
	    # First determine the true stack level the caller refers to.
	    set tclframe [frame2tclframe $frame_index]
	}
	{} {
	    # Called out of the "watched variables" window. Use the
	    # current stack frame to evaluate the property. This is
	    # similar to "locals", but without an explicit frame
	    # index.
	    set tclframe [frame2tclframe 0]
	}
	default {
	    log::log error "Illegal context \"$context\""
	    return -code error "Illegal context \"$context\""
	}
    }

    set properties {}

    if {
	[regexp {^\$\{(.*)\}$} $parent_expr -> aname] ||
	[regexp {^\$(.*)$}     $parent_expr -> aname] 
    } {
	# Get the array from tcl and convert its children into komodo properties
	# The array name is fully qualified if a global.

	log::log info "get children of array | $parent_expr | $aname"

	set data [dbg::getVar $tclframe -1 [list $aname]]
	set data [lindex $data 0]

	log::log info "array data = $data"

	set value {}
	foreach {name     type value} $data break ; # lassign
	#        == aname == a ^^^^^ 
	foreach {k v} $value {
	    lappend properties [xproperty $k [varquote ${aname}($k)] $v string 0]
	}
    } else {
	# Need namespace variables and subnamespaces.

	set nsName $parent_expr
	log::log info "get children of namespace | $parent_expr | $nsName"

	# Retrieve everything needed using fully qualified names and from the global scope.

	set subvars [GetResult [dbg::evaluate 0 [list ::info      vars    "${nsName}::*"]]]

	set subns   [GetResult [dbg::evaluate 0 [list ::namespace children $nsName]]]
	set subs    [dbg::getVar 0 -1 $subvars]

	foreach n $subns {lappend subs [list $n n {}]}

	set subs [lsort -ascii -index 0 $subs]

	log::log info "namespace data vars  = $subvars"
	log::log info "namespace data subns = $subns"
	log::log info "namespace data merge = $subs"

	set properties [cvtTclPropToPreXML $subs 0]

    }
    return [xstruct propertyInfo [xarray $properties]]
}

#######################################################
# Variables, namespaces for context and given stack frame

proc ::ktdb::protocol/Requester.getContext {arguments} {
    # Arguments = context/string, frame_index/int, radix/int
    # see getDebugPropertyInfo

    array set param $arguments
    validateParam context     string
    validateParam frame_index int
    validateParam radix       int

    switch -exact -- $context {
	globals {
	    set properties [getGlobalProperties]
	}
	locals  {
	    set properties [getLocalProperties $frame_index]
	}
	default {
	    log::log error "Illegal context \"$context\""
	    return -code error "Illegal context \"$context\""
	}
    }

    return [xstruct propertyInfo [xarray $properties]]
}

#######################################################

proc ::ktdb::getGlobalProperties {} {
    # Global properties are all global variables and all global namespaces.
    # The latter have children, again variables and namespaces ...
    # Array variables have children too.

    # level = #0, absolute global.
    # dbg::evaluate

    # (1) Get global variables
    # (2) Get global namespaces
    # (3) Merge and sort the two lists, then convert into list for consumption by the reply-generator.

    set globalVT  [dbg::getVariables 0]

    set globalV {}
    foreach g $globalVT {
	# Bugzilla #18701
	# Filter out internal variables of tcl and tk.

	set gname [lindex $g 0]
	if {[string match tcl_*  $gname]} {continue}
	if {[string match auto_* $gname]} {continue}
	if {[string match arg*   $gname]} {continue}
	if {[string match error* $gname]} {continue}
	if {[string match env    $gname]} {continue}
	if {[string match tk*    $gname]} {continue}

	lappend globalV ::$gname
    }

    set globals  [dbg::getVar 0 -1 $globalV]

    set globalNs [GetResult [dbg::evaluate 0 {::namespace children ::}]]
    foreach n $globalNs {lappend globals [list $n n {}]}

    set globals [lsort -ascii -index 0 $globals]

    log::log info "globals = $globals"

    return [cvtTclPropToPreXML $globals 0]
}

#######################################################

proc ::ktdb::getLocalProperties {frame} {
    set tclframe [frame2tclframe $frame]
    set filter 0

    if {$tclframe == 0} {
	set filter 1
    }

    ## We can add some more decisions here in the future.
    ## For example: locals in global namespace = {}
    ##              locals in other  namespace = namespace variables ...

    set localVT  [dbg::getVariables $tclframe]
    set localV {}

    foreach l $localVT {
	set lname [lindex $l 0]
	if {$filter} {
	    if {[string match tcl_*  $lname]} {continue}
	    if {[string match auto_* $lname]} {continue}
	    if {[string match arg*   $lname]} {continue}
	    if {[string match error* $lname]} {continue}
	    if {[string match env    $lname]} {continue}
	    if {[string match tk*    $lname]} {continue}
	}
	lappend localV $lname
    }

    set locals [lsort -ascii -index 0 [dbg::getVar $tclframe -1 $localV]]

    log::log info "locals = $locals"

    return [cvtTclPropToPreXML $locals $tclframe]
}

#######################################################

proc ::ktdb::varquote {vname} {
    if {
	![regexp {^[:A-Za-z0-9_]+$}       $vname] &&
	![regexp {^[:A-Za-z0-9_]+\(.*\)$} $vname]
    } {
	set vname "\{$vname\}"
    }
    return "\$$vname"
}

proc ::ktdb::cvtTclPropToPreXML {varval tclframe} {
    set properties {}
    foreach g $varval {
	foreach {name type value} $g break ; # lassign
	switch -exact -- $type {
	    s {
		set display [varquote [namespace tail $name]]
		set prop [xproperty $display [varquote $name] $value string 0]
	    }
	    a {
		# Arrays are transfered without values. We gave them
		# children, so lets Komodo ask about the contents if they are required.

		set display [varquote [namespace tail $name]]
		set children \
			[expr {[GetResult [dbg::evaluate $tclframe \
			[list array size $name]]] > 0}]

		set prop [xproperty $display [varquote $name] "{...}" array $children]
	    }
	    n {
		# Use only the last component of name as the screen name.
		# The full name is used when retrieving children, and that
		# contains the FQN name. Namespaces have no values, just
		# children

		if {
		    ([llength [GetResult [dbg::evaluate 0 [list ::info vars "${name}::*"]]]] > 0) ||
		    ([llength [GetResult [dbg::evaluate 0 [list ::namespace children $name]]]] > 0)
		} {
		    set children 1
		} else {
		    set children 0
		}

		set prop [xproperty ::[namespace tail $name] "$name" "{...}" namespace $children]
	    }
	    default {
		log::log critical "INTERNAL: Illegal type \"$type\""
		return -code error "INTERNAL: Illegal type \"$type\""
	    }
	}
	lappend properties $prop
    }
    return $properties
}

#######################################################
# What contexts are available ?

proc ::ktdb::protocol/Requester.getContextNamesAsString {arguments} {
    # Arguments =  frame_index/int
    variable contextNamesReply

    array set param $arguments
    validateParam frame_index int

    return $contextNamesReply
}

#######################################################

proc ::ktdb::protocol/Requester.addBreakpoint {arguments} {
    # Arguments = file/string lineno/int temporary/instance, cond/string

    variable bp
    variable bpcount
    variable bpTemp
    variable invalidBpReply

    ## The main work here is to transform the (file,lineno) combination
    ## from komodo into a 'location' as expected by the tcl engine.
    ## The other important work is to remember a mapping between tcl
    ## debugger bp id's and komodo id's
    ##
    ## Ignoring condition and temporary for now.

    array set param $arguments
    validateParam file      string
    validateParam lineno    int
    validateParam temporary int
    validateParam cond      int

    set fname [realname $file]

    # Use file and linenumber to get a location ...
    # Get block for the file -> Defined through instrumentation ...

    set fname   [fresolve $fname]
    set lines   [blk::getLines [set block [blk::makeBlock $fname]]]
    set nearest [dbg::binarySearch $lines $lineno]

    log::log debug "$fname => block $block (e[blk::exists $block], i[blk::isInstrumented $block]), lines = $lines"
    log::log debug "bp req. @ $lineno, nearest $nearest"


    if {$nearest !=  $lineno} {
	# Return -1 as indication of failure
	log::log warning "Invalid line for breakpoint"
	return $invalidBpReply
    }
    set location [loc::makeLocation $block $nearest]

    # We have something, tell the engine to create the bp ...

    set tclId [dbg::addLineBreakpoint $location]
    incr bpcount
    set bp($bpcount) $tclId

    ## No API to declare a created bp as temporary ...
    ## Tcl engine combines this with 'run' functionality
    ## Hack into the data structures ...

    if {$temporary} {
	set dbg::tempBreakPoint $tclId
	set bpTemp $bpcount
    }

    return [xstruct breakpointNumber [xint $bpcount]]
}

#######################################################

proc ::ktdb::protocol/Requester.setBreakpointAttribute {arguments} {
    # TODO, Arguments = ?
    return {}
}

#######################################################

proc ::ktdb::protocol/Requester.removeBreakpoint {arguments} {
    # Arguments = breakpointNumber/int

    variable bp

    array set param $arguments
    validateParam breakpointNumber int

    set tclId $bp($breakpointNumber)

    dbg::removeBreakpoint $tclId

    unset bp($breakpointNumber)

    return {}
}

#######################################################

###
### End of Protocol handlers.
###
#######################################################
#######################################################
#######################################################

proc ::ktdb::validateParam {name {type string}} {
    upvar param p
    if {![info exists p($name)]} {
	return -code error "Argument \"$name\" missing"
    }
    uplevel 1 [list set $name $p($name)]

    switch -exact -- $type {
	int {
	    if {![string is integer $p($name)]} {
		return -code error \
			"Expected integer for $name, got \"$p($name)\""
	    }
	}
	string {}
    }
    return
}

#######################################################

proc ::ktdb::frame2tclframe {frame_index} {
    # Tcl    Frame 0 = global
    # Komodo Frame 0 = Current
    #
    # IOW conversion required. Actually twice
    #
    # One, convert komodo frame into the tcl frame indexing into the
    # stack returned by getStack. Each frame there contains the level
    # it corresponds to in the true tcl stack (because of the virtual
    # frames). We get that value and then are able to ask the engine
    # about variables and stuff ...

    set dbgstack [dbg::getStack]
    set dbgframe [expr {[llength $dbgstack] - 1 - $frame_index}]
    set tclframe [lindex [lindex $dbgstack $dbgframe] 0]

    log::mlog debug [join $dbgstack \n] "* "
    log::log  info "FrameConvert $tclframe <- $frame_index \[[llength $dbgstack]\]| $dbgframe"

    return $tclframe
}


#######################################################
#######################################################
#######################################################
###
### Handlers for events coming from the engine and the nub
###

#######################################################

proc ::ktdb::traceAppState {args} {
    ::log::log debug "appState = $::dbg::appState"
}

#######################################################
log::log debug attachHandler

proc ::ktdb::attachHandler {args} {
    log::log notice "| engine attached to nub"
    logSystemState

    set ::forward 1
    return
}

#######################################################
log::log debug exitHandler

proc ::ktdb::exitHandler {} {
    log::log notice "| nub exit"
    log::log notice "suiciding server\\_"
    exit 0
    return
}

#######################################################
log::log debug errorHandler

proc ::ktdb::errorHandler {errMsg errStk errCode uncaught} {
    log::log  error "| nub error $errCode, $uncaught"
    log::log  error "| nub error $errMsg"
    log::mlog error $errStk "| nub error "

    logSystemState
    unlockRequests runTimeError

    # Drive the information about the error back to Komodo for display
    # in the output window.

    set   header "\[Stack trace\]"
    set   footer "__________________________"
    puts  stderr "$errMsg\n\n$header\n$footer\n$errStk\n$footer"
    flush stderr

    # FIXME\_ lowest common denominator right now, no post-mortem debugging => Kill ourselves and the child.
    dbg::quit
    #catch {kill $::dbg::appPid}
    exit 0
    # FIXME/  end

    return
}

#######################################################
log::log debug resultHandler

proc ::ktdb::resultHandler {id code result errCode errInfo} {
    # Assume TCL_OK for now ...
    logSystemState

    variable eCodeV  $errCode
    variable eInfoV  $errInfo
    variable codeV   $code
    variable resultV $result
    variable idV     $id ; # unlock the vwait ...
    return
}

proc ::ktdb::GetResult {id} {
    variable resultV
    variable eCodeV
    variable eInfoV
    variable codeV

    vwait ::ktdb::idV

    if {$codeV != 0} {
	return -code $codeV -errorinfo $eInfoV -errorcode $eCodeV $resultV
    }
    return  $resultV
}

#######################################################
log::log debug *breakHandler

proc ::ktdb::linebreakHandler {args} {
    logSystemState
    cleanTempBp
    unlockRequests lineBreak
}

proc ::ktdb::varbreakHandler {var type} {
    logSystemState
    cleanTempBp
    unlockRequests varBreak
}

proc ::ktdb::userbreakHandler {args} {
    logSystemState
    cleanTempBp
    unlockRequests userBreak
}

proc ::ktdb::cleanTempBp {} {
    variable bp
    variable bpTemp

    if {$bpTemp != {} && $dbg::tempBreakPoint == {}} {
	# Remove mapping, implicit removeBreakpoint
	unset bp($bpTemp)
	set bpTemp {}
    }
    return
}

proc ::ktdb::unlockRequests {why} {
    set ::ktdb::stepComplete $why
    log::log debug "unlock - $why <- [info level 1]"
}

#######################################################

###
### End of Handlers for events coming from the engine and the nub
###
#######################################################
#######################################################
#######################################################

#######################################################
#
# xmlrpc helpers to generate proper reply lists.
#

proc ::ktdb::xquote {value} {
    array set Entity {
	$ $
	< &lt;
	> &gt;
	& &amp;
	\" &quot;
	' &apos;
    }

    regsub -all {([$<>&"'])} $value {$Entity(\1)} value
    return [subst -nocommand -nobackslash $value]
}

proc ::ktdb::xstring {text}  {return <value><string>[xquote $text]</string></value>}
proc ::ktdb::xint    {i}     {return <value><i4>[xquote $i]</i4></value>}
proc ::ktdb::xbool   {b}     {return <value><boolean>[xquote $b]</boolean></value>}

proc ::ktdb::xstruct {args}  {
    set xml <value><struct>
    foreach {k v} $args {
	append xml <member><name>$k</name>$v</member>
    }
    return $xml</struct></value>
}

proc ::ktdb::xarray  {items} {
    return <value><array><data>[join $items ""]</data></array></value>
}

proc ::ktdb::xnamestruct {name full} {
    xstruct name [xstring $name] fullName [xstring $full]
}

proc ::ktdb::xproperty {name full value type children} {
    xstruct name  [xnamestruct $name $full] \
	    value [xstring $value] \
	    type  [xstring $type] \
	    hasChildren [xbool $children]
}

proc ::ktdb::uriname {name} {
    set pre file://
    if {$::tcl_platform(platform) == "windows"} { append pre / }
    return $pre$name
}
proc ::ktdb::realname {name} {
    set pre file://
    if {$::tcl_platform(platform) == "windows"} { append pre / }
    regsub ^$pre $name {} name

    # Bugzilla entry #17777
    # Handle encoded characters in the url. (Like %20 = space)
    # Minimal fix, only spaces are handled.

    set name [string map {%20 { }} $name]

    return $name
}

# Initialize some constant strings now ...
namespace eval ::ktdb {
    set statusReply        [xstruct status [xstring hello]]
    set contextNamesReply  [xstruct names  [xstring {globals locals}]]
    set invalidBpReply     [xstruct breakpointNumber [xint -1]]
    set setDebugOkReply    [xstruct setResultString [xstring ""]]
    set virtualGlobalFrame [xstruct source [xstring {# Speciality of the tcl debugger:
# * The script has entered the event loop and is therefore in the C-level of
#   Tcl. From now on the tcl debugger has control only if an event causes the
#   execution of a tcl script and this script contains one or more breakpoints.
#
# * While the event loop is in control Komodo is unable to query the application
#   about stack, variables, etc.
#
# Note:
# * If a window of the application does not show its true contents but is empty
#   the event loop is not in complete control. Please use F5 (Continue) to
#   unlock it.
}]]
}

#
# End of xmlrpc helpers to generate proper reply lists.
#
#######################################################

#######################################################
# Resolve a filename to something without symlinks and such

proc ::ktdb::fresolve {fname} {
    if {![file exists $fname]} {
	return -code error "file $fname does not exist"
    }
    set path [file dirname $fname]
    set here [pwd]
    cd $path
    set path [pwd]
    cd $here
    return [file join $path [file tail $fname]]
}

#######################################################

::ktdb::main
exit 0
